<?php

require_once( dirname(__FILE__)."/ToolsForPHP.php" );
require_once( dirname(__FILE__)."/QRInformation.php" );
require_once( dirname(__FILE__)."/QRDataElement.php" );
require_once( dirname(__FILE__)."/QRReadSolomon.php" );
             

class CQRMatrix
{
    protected $m_listQRData;
    protected $m_ppucQRMatric;
    protected $m_ppucMaskMatrix;
    protected $m_iModuleNumber;
    protected $m_iModelTypeIndex;
    protected $m_iDataBitNumber;
    protected $m_iDataSizeNumber;
    protected $m_pucData;
    protected $m_iErrorSizeNumber;
    protected $m_pucError;
    protected $m_iQRMaskPattern;
    
    public function __construct()
    {
        $this->m_listQRData = array();
        $this->m_iModelTypeIndex = -1;
        $this->m_iModuleNumber = 0;
        $this->m_ppucQRMatrix = array();
        $this->m_ppucMaskMatrix = array();

        $this->m_iDataBitNumber = 0;
        $this->m_iDataSizeNumber = 0;
        $this->m_pucData = "";

        $this->m_iErrorSizeNumber = 0;
        $this->m_pucError = "";

        $this->m_iQRMaskPattern = -1;
    }
    
    public function __destruct()
    {
        $this->ReleaseAll();
    }
    
    ///////////////////////////////////////////////////////////
    ///		内部リストに文字列データをセットする
    ///////////////////////////////////////////////////////////
    public function AddQRCharacters( $pcszString , $iCharacterType )
    {
        $bDataAdded = FALSE;
        
        //１つ以上のデータが登録されている場合
        if( count( $this->m_listQRData ) != 0 )
        {
            $iLastIndex = count( $this->m_listQRData ) - 1;
            
            $pQRElement = $this->m_listQRData[ $iLastIndex ];
            
            //最後のデータモードと現在のデータモードが一致している場合には最後のリストに
            //文字列を追加する
            $bDataAdded = $pQRElement->AddData( $pcszString, strlen($pcszString), $iCharacterType );
        }
        
        //データが追加されなかった場合
        if( !$bDataAdded )
        {
            $qrElement = new CQRDataElement();
            $qrElement->SetData( $pcszString, strlen($pcszString), $iCharacterType );
            $this->m_listQRData[] = $qrElement;
        }
        
        /* ForDebug
        var_dump( $this->m_listQRData );
        //*/
        
        return TRUE;
    }

    ///////////////////////////////////////////////////////////
    ///		セットされているデータを元にＱＲマトリックスを作成する
    ///		引数：
    ///			iCollectLevel	：	誤り訂正レベル
    ///									COLLECTLEVEL_L
    ///									COLLECTLEVEL_M
    ///									COLLECTLEVEL_Q
    ///									COLLECTLEVEL_H
    ///								のいずれかを指定する
    ///		戻り値：
    ///			==false：作成に失敗（データが大きすぎるなど）
    ///			==true：作成に成功
    ///////////////////////////////////////////////////////////
    //bool MakeQRMatrix( int iCollectLevel , int & iErrorType );
    public function MakeQRMatrix( $iCollectLevel, &$iErrorType )
    {
        $this->ReleaseAll();
        
        $bRet = TRUE;
        $iErrorType = QRERR_NOERROR;

        //型番情報を設定する
        $bRet = $this->SetTypeNumber( $iCollectLevel, $iErrorType );

        //データビット列を作成する
        if( $bRet )
        {
            $bRet = $this->CreateDataBits( $iErrorType );
        }

        //誤り訂正コード文字列の作成
        if( $bRet )
        {
            $bRet = $this->CreateErrorCollectData( $iErrorType );
        }

        //セットされた型番号から、モジュール配列を初期化する
        if( $bRet )
        {
            $bRet = $this->InitModuleArea( $iErrorType );
        }

        //位置検出パターンのセット
        if( $bRet )
        {
            $bRet = $this->SetPositionDitectPattern( $iErrorType );
        }

        //位置あわせパターンのセット
        if( $bRet )
        {
            $bRet = $this->SetPositionCheckPattern( $iErrorType );
        }

        //タイミングパターンのセット
        if( $bRet )
        {
            $bRet = $this->SetTimingPattern( $iErrorType );
        }

        //形式情報の仮出力
        if( $bRet )
        {
            $bRet = $this->SetTypeInformation( TRUE , $iErrorType );
        }

        //型番情報
        if( $bRet )
        {
            $bRet = $this->SetModelInformation( $iErrorType );
        }

        //マスクパターン適応領域を取得する
        if( $bRet )
        {
            $bRet = $this->CreateMaskPatternArea( $iErrorType );
        }

        //データビットの出力
        if( $bRet )
        {
            $bRet = $this->OutputDataBits( $iErrorType );
        }

        //マスクパターンの決定
        if( $bRet )
        {
            $bRet = $this->DecideMaskPattern( $iErrorType );
        }

        //形式情報の出力
        if( $bRet )
        {
            $bRet = $this->SetTypeInformation( FALSE, $iErrorType );
        }//*/

        return $bRet;
    }
	///////////////////////////////////////////////////////////
	///		QRコードの型数を取得する
	///		型数は1～40となる
	///////////////////////////////////////////////////////////
	public function GetTypeNumber()
	{
		return $this->m_iModelTypeIndex;
	}
	///////////////////////////////////////////////////////////
	///		モジュール個数を取得する
	///		モジュール個数がマトリックスの縦・横数となる
	///////////////////////////////////////////////////////////
	public function GetModuleNumber()
	{
		return $this->m_iModuleNumber;
	}
    
    ///////////////////////////////////////////////////////////
	///		指定された行列に対応するポイントデータを取得する
	///		戻り値：　==false：明領域
	///				　==true：暗領域
	///////////////////////////////////////////////////////////
	public function IsDarkPoint( $iColumnIndex , $iRowIndex )
	{
		if( $iColumnIndex < 0 || $iColumnIndex >= $this->m_iModuleNumber ||
			$iRowIndex    < 0 || $iRowIndex    >= $this->m_iModuleNumber )
		{
			return FALSE;
		}

		$iModulePointOrg = 0;

		//未処理領域の場合
		if( $this->m_ppucQRMatrix[$iColumnIndex][$iRowIndex] < 0 )
		{
			$iModulePointOrg = 0;
		}
		//処理済み領域の場合
		else
		{
			$iModulePointOrg = $this->GetModulePoint( $iColumnIndex , $iRowIndex ) % 2;
		}

		//マスクがかからないデータの場合
		if( $this->m_ppucMaskMatrix[$iColumnIndex][$iRowIndex] == QR_UNKNOWN || $this->m_iQRMaskPattern < 0 )
		{
			return ($iModulePointOrg !=0 );
		}

		//マスクがかかるデータの場合
		return ((($this->GetMaskPoint( $iColumnIndex , $iRowIndex ) % 2) ^ $iModulePointOrg) != 0 );
	}
    
    ///////////////////////////////////////////////////////////
    ///		全データを削除する
    ///////////////////////////////////////////////////////////
    protected function ReleaseAll()
    {
        $this->ReleaseQRMatrix();
        $this->ReleaseMaskMatrix();
        $this->ReleaseData();
        $this->ReleaseError();

        $this->m_iModuleNumber = 0;
        $this->m_iModelTypeIndex = -1;
    }
    
    ///////////////////////////////////////////////////////////
    ///		QRのデータを削除する
    ///////////////////////////////////////////////////////////
    //void ReleaseQRMatrix();
    protected function ReleaseQRMatrix()
    {
        $this->m_ppucQRMatrix = array();
    }

    ///////////////////////////////////////////////////////////
    ///		マスクのデータを削除する
    ///////////////////////////////////////////////////////////
	//void ReleaseMaskMatrix();
    protected function ReleaseMaskMatrix()
    {
        $this->m_ppucMaskMatrix = array();
    }
    
	//void ReleaseData();
    protected function ReleaseData()
    {
        $this->m_pucData = "";
        $this->m_iDataSizeNumber = 0;
    }
    
	//void ReleaseError();
    protected function ReleaseError()
    {
        $this->m_pucError = "";
        $this->m_iErrorSizeNumber = 0;
    }
    
    ///////////////////////////////////////////////////////////
    ///		登録済みのデータリストからデータのビット数を取得する
    ///		ＱＲの型番により、サイズが異なるため、各型番時のデータサイズを
    ///		セットする
    ///		引数：
    ///			iDataBit1_9		(OUT)	1型～9型のデータビット数
    ///			iDataBit10_26	(OUT)	10型～26型のデータビット数
    ///			iDataBit27_40	(OUT)	27型～40型のデータビット数
    ///////////////////////////////////////////////////////////
    //bool GetDataBitNumber( int & iDataBit1_9 , int & iDataBit10_26 , int & iDataBit27_40 );
    protected function GetDataBitNumber( &$iDataBit1_9 , &$iDataBit10_26 , &$iDataBit27_40 )
    {
        $bRet = TRUE;

        //初期化
        $iDataBit1_9 = $iDataBit10_26 = $iDataBit27_40 = 0;
        //型に関係ないデータビット数
        $iDataBits = 0;

        $iListLength = count( $this->m_listQRData );
        for( $iListIndex=0
            ; $iListIndex<$iListLength
            ; $iListIndex++ )
        {
            $pDataElement = $this->m_listQRData[$iListIndex];
            
            //データ指示子
            $iDataBits += 4;

            switch( $pDataElement->GetDataMode() )
            {
            //数字モードの場合
            case QRCHARACTERTYPE_NUMERIC:

                //データサイズ出力用
                $iDataBit1_9   += 10;
                $iDataBit10_26 += 12;
                $iDataBit27_40 += 14;

                //データ本体出力用
                $iDataBits += floor( $pDataElement->GetDataLength() / 3 ) * 10;
                if( $pDataElement->GetDataLength() % 3 == 1 )
                {
                    $iDataBits += 4;
                }
                else if( $pDataElement->GetDataLength() % 3 == 2 )
                {
                    $iDataBits += 7;
                }
                break;

            //アルファベットモードの場合
            case QRCHARACTERTYPE_ALPHABET:

                //データサイズ出力用
                $iDataBit1_9   += 9;
                $iDataBit10_26 += 11;
                $iDataBit27_40 += 13;

                //データ本体出力用
                $iDataBits += floor( $pDataElement->GetDataLength() / 2 ) * 11;
                if( $pDataElement->GetDataLength() % 2 ) 
                {
                    $iDataBits += 6;
                }

                break;

            //ASCII文字列
            case QRCHARACTERTYPE_ASCII:

                //データサイズ出力用
                $iDataBit1_9 += 8;
                $iDataBit10_26 += 16;
                $iDataBit27_40 += 16;

                $iDataBits += $pDataElement->GetDataLength() * 8;
                break;

            //漢字文字列用
            case QRCHARACTERTYPE_KANJI:

                //データサイズ出力用
                $iDataBit1_9   += 8;
                $iDataBit10_26 += 10;
                $iDataBit27_40 += 12;

                //データ出力用
                $iDataBits += floor($pDataElement->GetDataLength() / 2) * 13;

                break;

            default:
                return FALSE;
            }
        }

        //終端モード指示子
        $iDataBits += 4;

        $iDataBit1_9 += $iDataBits;
        $iDataBit10_26 += $iDataBits;
        $iDataBit27_40 += $iDataBits;

        return $bRet;
    }
    
    //bool CreateDataBits( int & iErrorType );
    ///////////////////////////////////////////////////////////
    ///		登録されているデータからデータ文字列を作成する
    ///////////////////////////////////////////////////////////
    protected function CreateDataBits( &$iErrorType )
    {
        global $c_typeMatrix;
        
        $bRet = TRUE;

        //データのリリース
        $this->ReleaseData();

        //list<CQRDataElement>::iterator pDataElement;
        $iListLength = count( $this->m_listQRData );
        
        for( $iListIndex = 0
            ;$iListIndex < $iListLength && $bRet 
            ;$iListIndex++ )
        {
            $pDataElement = $this->m_listQRData[$iListIndex];
            
            //データモードの出力
            $cModel = 0x00;
            switch( $pDataElement->GetDataMode() )
            {
            case QRCHARACTERTYPE_NUMERIC:
                $bRet = $this->AddDataBits_Numeric( $pDataElement->GetData() , $pDataElement->GetDataLength() , $iErrorType);
                break;

            case QRCHARACTERTYPE_ALPHABET:
                $bRet = $this->AddDataBits_Alphabet( $pDataElement->GetData() , $pDataElement->GetDataLength() , $iErrorType );
                break;

            case QRCHARACTERTYPE_ASCII:
                $bRet = $this->AddDataBits_Ascii( $pDataElement->GetData() , $pDataElement->GetDataLength() , $iErrorType );
                break;

            case QRCHARACTERTYPE_KANJI:
                $bRet = $this->AddDataBits_Kanji( $pDataElement->GetData() , $pDataElement->GetDataLength() , $iErrorType );
                break;
            }

        }


        //終端出力
        if( $bRet )
        {
            $bRet = $this->AddDataBits( 0x00 , 4 , $iErrorType);
        }

        if( $bRet )
        {
            //データバイト終端までを0で埋める
            for( ; $this->m_iDataBitNumber < $c_typeMatrix[$this->m_iModelTypeIndex]->m_iDataBitNumber &&
                   $this->m_iDataBitNumber % 8 != 0 && $bRet
                 ; )
            {
                $bRet = $this->AddDataBits( 0x00 , 1 , $iErrorType );
            }

            $bFirstByte = TRUE;
            for( ; $this->m_iDataBitNumber < $c_typeMatrix[$this->m_iModelTypeIndex]->m_iDataBitNumber && $bRet
                 ;  )
            {
                if( $bFirstByte )
                {
                    $bRet = $this->AddDataBits( 0xec , 8 , $iErrorType );
                }
                else
                {
                    $bRet = $this->AddDataBits( 0x11 , 8 , $iErrorType );
                }
                $bFirstByte = !$bFirstByte;
            }
        }

        return $bRet;
    }
    
    
    ///////////////////////////////////////////////////////////
    ///		指定されたビット番号のビットマスクを取得する
    ///////////////////////////////////////////////////////////
    //unsigned char GetBitMask( int iBitNumber );
    protected function GetBitMask( $iBitNumber )
    {
        $ucMask = (0x01 << $iBitNumber);
        /*switch( $iBitNumber )
        {
            case 0: $ucMask = "\x01";   break;
            case 1: $ucMask = "\x02";   break;
            case 2: $ucMask = "\x04";   break;
            case 3: $ucMask = "\x08";   break;
            case 4: $ucMask = "0x10";   break;
            case 5: $ucMask = "0x20";   break;
            case 6: $ucMask = "0x40";   break;
            case 7: $ucMask = "0x80";   break;
            default:    $ucMask = "0x00";
        }*/

        return $ucMask;
    }    
    
    ///////////////////////////////////////////////////////////
    ///		データのバイト数と誤り訂正レベルから
    ///		QRの型（1型～40型）を決定する
    ///////////////////////////////////////////////////////////
    //bool SetTypeNumber(int iCollectLevel , int & iErrorType );
    protected function SetTypeNumber($iCollectLevel , &$iErrorType )
    {
        global $c_typeMatrix;
        
        $iByteNum = 0;

        //指定された誤り訂正レベルが不正の場合
        if( $iCollectLevel != COLLECTLEVEL_L && $iCollectLevel != COLLECTLEVEL_M &&
            $iCollectLevel != COLLECTLEVEL_Q && $iCollectLevel != COLLECTLEVEL_H )
        {
            $iErrorType = QRERR_WRONG_COLLECTTYPE;
            return FALSE;
        }

        $bRet = FALSE;

        if( !$this->GetDataBitNumber( $iDataBit1_9 , $iDataBit10_26 , $iDataBit27_40 ) )
        {
            $iErrorType = QRERR_DATABITCOUNT_FAILURE;
            return FALSE;
        }

        //型番号を情報構造体から検索する
        for( $i=0 ; $c_typeMatrix[$i]->m_iModelNumber > 0 ; $i++ )
        {
            //型によってサイズが異なるため、現在のチェック型に合わせたビットサイズを取得する
            $iDataBitNumber = $iDataBit1_9;
            if( $c_typeMatrix[$i]->m_iModelNumber >= 27 )
            {
                $iDataBitNumber = $iDataBit27_40;
            }
            else if( $c_typeMatrix[$i]->m_iModelNumber >= 10 )
            {
                $iDataBitNumber = $iDataBit10_26;
            }

            if( $c_typeMatrix[$i]->m_iCollectLevel == $iCollectLevel &&
                $c_typeMatrix[$i]->m_iDataBitNumber >= $iDataBitNumber )
            {
                $this->m_iModelTypeIndex = $i;

                $bRet = TRUE;
                break;
            }
        }

        //エラーが発生した場合
        if( !$bRet )
        {
            $iErrorType = QRERR_DATASIZE_LARGE;
        }

        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		設定された型番号を使用して、モジュール領域を
    ///		確保し、初期化する
    ///////////////////////////////////////////////////////////
    //bool InitModuleArea(int & iErrorType);
    protected function InitModuleArea( &$iErrorType )
    {
        global $c_typeMatrix;
        
        //型番号からモジュール個数を計算
        $this->m_iModuleNumber = 17 + $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber * 4;

        //モジュールデータの領域確保と初期化
        $this->m_ppucQRMatrix = array();
        //マスクデータの領域確保と初期化
        $this->m_ppucMaskMatrix = array();

        for( $i=0 ; $i<$this->m_iModuleNumber ; $i++ )
        {
            $this->m_ppucQRMatrix[$i] = array();
            $this->m_ppucMaskMatrix[$i] = array();

            for( $j=0 ; $j<$this->m_iModuleNumber ; $j++ )
            {
                $this->m_ppucQRMatrix[$i][$j] = QR_UNKNOWN;
                $this->m_ppucMaskMatrix[$i][$j] = QR_UNKNOWN;
            }
        }

        return TRUE;
    }
    
    ///////////////////////////////////////////////////////////
    ///		位置検出パターンのセット
    ///		位置検出パターンの分離パターンも同時にセットする
    ///////////////////////////////////////////////////////////
    //bool SetPositionDitectPattern( int & iErrorType );
    protected function SetPositionDitectPattern( &$iErrorType )
    {
        $bRet = TRUE;

        $cPatterns = array();
        $cPatterns[] = "000000000";
        $cPatterns[] = "011111110";
        $cPatterns[] = "010000010";
        $cPatterns[] = "010111010";
        $cPatterns[] = "010111010";
        $cPatterns[] = "010111010";
        $cPatterns[] = "010000010";
        $cPatterns[] = "011111110";
        $cPatterns[] = "000000000";

        //一検出パターン位置
        $iStartPosX = array( -1 , $this->m_iModuleNumber - 8 , -1 );
        $iStartPosY = array( -1 , -1, $this->m_iModuleNumber - 8 );

        for( $i=0 ; $i<3 ; $i++ )
        {
            //分離パターンの出力
            for( $j=0 ; $j<9 ; $j++ )
            {
                for( $k=0 ; $k<9 ; $k++ )
                {
                    $iModuleType = QR_POSITIONDITECT_LIGHT;

                    if( $cPatterns[$j][$k] == "1" )
                    {
                        $iModuleType = QR_POSITIONDITECT_DARK;
                    }

                    $x = $iStartPosX[$i] + $j;
                    $y = $iStartPosY[$i] + $k;

                    $this->SetModulePoint( $x , $y , $iModuleType );
                }
            }

        }

        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		位置あわせパターンのセット
    ///////////////////////////////////////////////////////////
	//bool SetPositionCheckPattern( int & iErrorType );
    protected function SetPositionCheckPattern( &$iErrorType )
    {
        global $c_typeMatrix;
        global $c_sPositionChecks;
        
        $bRet = TRUE;

        $cPatterns = array();
        $cPatterns[] = "11111";
        $cPatterns[] = "10001";
        $cPatterns[] = "10101";
        $cPatterns[] = "10001";
        $cPatterns[] = "11111";

        //型番による位置あわせパターンの縦横方向ループ
        $iModelNumber = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber;
        
        for( $i=0 ; $c_sPositionChecks[$iModelNumber-1][$i] ; $i++ )
        {
            for( $j=0 ; $j<$c_sPositionChecks[$iModelNumber-1][$j] ; $j++ )
            {
                //位置検出パターンとの被りがないかどうかをチェックする
                $bOutput = TRUE;

                //左上隅
                $x = $c_sPositionChecks[$iModelNumber-1][$i] - 2;
                $y = $c_sPositionChecks[$iModelNumber-1][$j] - 2;
                if( $this->GetModulePoint( $x , $y ) != QR_UNKNOWN )
                {
                    $bOutput = FALSE;
                }

                //右上隅
                $x += 4;
                if( $this->GetModulePoint( $x , $y ) != QR_UNKNOWN )
                {
                    $bOutput = FALSE;
                }

                //右下隅
                $y += 4;
                if( $this->GetModulePoint( $x , $y ) != QR_UNKNOWN )
                {
                    $bOutput = FALSE;
                }

                //左下隅
                $x -= 4;
                if( $this->GetModulePoint( $x , $y ) != QR_UNKNOWN )
                {
                    $bOutput = FALSE;
                }

                if( $bOutput )
                {
                    //各位置あわせパターンの出力用ボックスループ
                    for( $k = 0 ; $k < 5 ; $k++ )
                    {
                        for( $l = 0 ; $l < 5 ; $l++ )
                        {
                            $x = $c_sPositionChecks[$iModelNumber-1][$i] + $k - 2;
                            $y = $c_sPositionChecks[$iModelNumber-1][$j] + $l - 2;

                            //黒か白かの判定
                            $iModuleType = QR_POSITIONCHECK_LIGHT;
                            if( $cPatterns[$k][$l] == "1" )
                            {
                                $iModuleType = QR_POSITIONCHECK_DARK;
                            }
                            $this->SetModulePoint( $x , $y , $iModuleType );

                        }
                    }
                }
            }
        }

        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		タイミングパターンのセット
    ///////////////////////////////////////////////////////////
    //bool SetTimingPattern( int & iErrorType );
    protected function SetTimingPattern( &$iErrorType )
    {
        $bRet = TRUE;

        for( $i=0 ; $i<$this->m_iModuleNumber ; $i++ )
        {
            $iModuleType = (($i%2)?(QR_TIMING_LIGHT):(QR_TIMING_DARK));

            $x = $i;
            $y = 6;
            if( $this->GetModulePoint( $x , $y ) == QR_UNKNOWN )
            {
                $this->SetModulePoint( $x , $y , $iModuleType );
            }

            $x = 6;
            $y = $i;
            if( $this->GetModulePoint( $x , $y ) == QR_UNKNOWN )
            {
                $this->SetModulePoint( $x , $y , $iModuleType );
            }

        }

        return $bRet;
    }

    ///////////////////////////////////////////////////////////
    ///		形式情報を書き出す
    ///		引数：
    ///			bReserveOnly	：	trueを指定するとデータ領域
    ///								の予約を行う（すべてLIGHTで塗りつぶす）
    ///////////////////////////////////////////////////////////
    //bool SetTypeInformation( bool bReserveOnly , int & iErrorType );
    protected function SetTypeInformation( $bReserveOnly , &$iErrorType )
    {
        global $c_typeMatrix;
        
        //予約のみの場合
        if( $bReserveOnly )
        {
            //左上の横方向
            for( $i=0 ; $i<9 ; $i++ )
            {
                $x = $i;
                $y = 8;
                if( $this->GetModulePoint( $x , $y ) == QR_UNKNOWN )
                {
                    $this->SetModulePoint( $x , $y , QR_TYPE_LIGHT );
                }
            }

            //左上の縦方向
            for( $i = 0 ; $i < 8 ; $i++ )
            {
                $x = 8;
                $y = $i;
                if( $this->GetModulePoint( $x , $y ) == QR_UNKNOWN )
                {
                    $this->SetModulePoint( $x , $y , QR_TYPE_LIGHT );
                }
            }

            //右上の横方向
            for( $i=0 ; $i<8 ; $i++ )
            {
                $x = $this->m_iModuleNumber - 8 + $i;
                $y = 8;
                if( $this->GetModulePoint( $x , $y ) == QR_UNKNOWN )
                {
                    $this->SetModulePoint( $x , $y , QR_TYPE_LIGHT );
                }
            }

            //左下の縦方向
            for( $i=0 ; $i<8 ; $i++ )
            {
                $x = 8;
                $y = $this->m_iModuleNumber - 8 + $i;
                if( $this->GetModulePoint( $x , $y ) == QR_UNKNOWN )
                {
                    $this->SetModulePoint( $x , $y , QR_TYPE_LIGHT );
                }
            }

            return TRUE;
        }

        //本番出力
        $ucDataBits= "000000000000000";

        //エラー訂正符号の取得

        $ucTypeData = $this->m_iQRMaskPattern & 0x07;
        $ucTypeData = $ucTypeData | ((($c_typeMatrix[$this->m_iModelTypeIndex]->m_iCollectLevel) << 3) & 0x18);

        //bTypeBitsの頭５桁にセットする
        for( $i=0 ; $i<5 ; $i++ )
        {
            $ucMask = $this->GetBitMask( 4 - $i );
            $ucDataBits[$i] = (( ($ucMask & $ucTypeData ) != 0 )?("1"):("0"));
        }

        $ucCoefficients = "10100110111";

        $bRet = FALSE;

        //エラービットを追加する
        $bRet = $this->GetExpressionMod( $ucDataBits , 5 , $ucCoefficients , 5 , 10 );

        //マスクパターンの適用
        $ucMask = "101010000010010";

        for( $i=0 ; $i<15 ; $i++ )
        {
            if( $ucDataBits[$i] == $ucMask[$i] )
            {
                $ucDataBits[$i] = "0";
            }
            else
            {
                $ucDataBits[$i] = "1";
            }
        }

        //出力処理

        //左上横
        $iTemp = 0;
        for( $i = 0 ; $i < 8 ; $i++,$iTemp++ )
        {
            //タイミングパターンに重なる場合
            if( $i == 6 )
            {
                $iTemp++;
            }

            $iPointData = QR_TYPE_DARK;
            if( $ucDataBits[$i] == "0" )
            {
                $iPointData = QR_TYPE_LIGHT;
            }
            $this->SetModulePoint( $iTemp , 8 , $iPointData );
        }

        //左上縦
        $iTemp = 0;
        for( $i = 0 ; $i < 7 ; $i++,$iTemp++ )
        {
            //タイミングパターンに重なる場合
            if( $i == 6 )
            {
                $iTemp++;
            }

            $iPointData = QR_TYPE_DARK;
            if( $ucDataBits[14-$i] == "0" )
            {
                $iPointData = QR_TYPE_LIGHT;
            }
            $this->SetModulePoint( 8 , $iTemp , $iPointData );
        }

        //左下
        for( $i = 0 ; $i < 8 ; $i++ )
        {
            $iPointData = QR_TYPE_DARK;
            if( $i != 7 && $ucDataBits[$i] == "0" )
            {
                $iPointData = QR_TYPE_LIGHT;
            }

            $this->SetModulePoint( 8 , $this->m_iModuleNumber - $i - 1 , $iPointData );
        }

        //右上
        for( $i = 0 ; $i < 8 ; $i++ )
        {
            $iPointData = QR_TYPE_DARK;
            if( $ucDataBits[14-$i] == '0' )
            {
                $iPointData = QR_TYPE_LIGHT;
            }

            $this->SetModulePoint( $this->m_iModuleNumber - $i - 1 , 8 , $iPointData );
        }

        return $bRet;
    }

    ///////////////////////////////////////////////////////////
    ///		型番情報を書き込む
    ///////////////////////////////////////////////////////////
    //bool SetModelInformation( int & iErrorType );
    protected function SetModelInformation( &$iErrorType )
    {
        global $c_typeMatrix;
        
        //型番情報は7型未満の場合（型番情報を記述しない）
        if( $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber < 7 )
        {
            return TRUE;
        }

        $ucDataBits = "000000000000000000";
        $ucModel = ( $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber) & 0x3f;

        for( $i=0 ; $i<6 ; $i++ )
        {
            $ucMask = $this->GetBitMask( 5 - i );
            if( ( $ucModel & $ucMask ) != 0 )
            {
                $ucDataBits[$i] = '1';
            }
        }

        $ucCoefficients = "1111100100101";
        $bRet = $this->GetExpressionMod( $ucDataBits , 6 , $ucCoefficients , 6 , 12 );

        //データの出力
        for( $i=0 ; $i<6 ; $i++ )
        {
            for( $j=0 ; $j<3 ; $j++ )
            {
                $iPointData = QR_MODEL_DARK;
                if( $ucDataBits[(5 - $i) * 3 + (2 - $j)] == '0' )
                {
                    $iPointData = QR_MODEL_LIGHT;
                }

                //左側の型番データを出力
                $x = $i;
                $y = $j + $this->m_iModuleNumber - 11;
                $this->SetModulePoint( $x , $y , $iPointData );

                //右側の型番データを出力
                $x = $j + $this->m_iModuleNumber - 11;
                $y = $i;
                $this->SetModulePoint( $x , $y , $iPointData );
            }
        }

        return TRUE;
    }
    
    ///////////////////////////////////////////////////////////
    ///		マスクパターンの適応領域の取得
    ///		現在のモジュール領域内で、QR_UNKNOWNでない部分を
    ///		マスクパターン適応領域としてセットする
    ///////////////////////////////////////////////////////////
	//bool CreateMaskPatternArea( int & iErrorType );
    protected function CreateMaskPatternArea( &$iErrorType )
    {
        for( $i=0 ; $i<$this->m_iModuleNumber ; $i++ )
        {
            for( $j=0 ; $j<$this->m_iModuleNumber ; $j++ )
            {
                if( $this->m_ppucQRMatrix[$i][$j] == QR_UNKNOWN )
                {
                    $this->m_ppucMaskMatrix[$i][$j] = QR_MASK_LIGHT;
                }
            }
        }

        return TRUE;
    }
    
	//bool AddDataBits( unsigned char ucData , int iBitNumber , int & iErrorType );
    ///////////////////////////////////////////////////////////
    ///		データにビットを追加する
    ///////////////////////////////////////////////////////////
	//bool AddDataBits( unsigned char ucData , int iBitNumber , int & iErrorType );
    protected function AddDataBits( $ucData /*数値型で送信*/ , $iBitNumber , &$iErrorType )
    {
        /* For Debug
        printf( "AddDataBits..ucData=%02X, iBitNumber=%d<br>\n", $ucData, $iBitNumber );
        //*/
        
        $bRet = TRUE;

        if( $iBitNumber <= 0 || $iBitNumber > 8 )
        {
            $iErrorType = QRERR_WRONGDIRECT_BIT;
            return false;
        }

        $iBitLast = $this->m_iDataBitNumber + $iBitNumber;
        //必要バイトサイズ
        $iByteNow = floor(($this->m_iDataBitNumber + 7) / 8);
        $iByteNeed = floor( ($iBitLast + 7) / 8 );

        //追加のデータ領域が必要な場合
        if( $iByteNeed > $this->m_iDataSizeNumber )
        {
            //128バイト
            $this->m_iDataSizeNumber += 128;
            
            memset( $puszTemp , "\x00" , $this->m_iDataSizeNumber );

            if( strlen( $this->m_pucData ) != 0 )
            {
                memcpy( $puszTemp , $this->m_pucData , $iByteNow );
            }

            $this->m_pucData = $puszTemp;
        }

        //今からセットするデータのマスク
        $ucSetMask = $this->GetBitMask( $iBitNumber - 1 );

        //データの追加処理
        for( $i=0 ; $i < $iBitNumber ; $i++ )
        {
            //bitがonの場合
            if( ($ucSetMask & $ucData) != 0 )
            {
                $iBitToSet = $this->m_iDataBitNumber + $i;

                //セット済みデータの次のデータのマスク
                $ucDataMask = $this->GetBitMask( 7 - ($iBitToSet % 8) );

                //該当データのbitもonにする
                $iByteNow = $iBitToSet / 8;

                $this->m_pucData[$iByteNow] = chr( ord($this->m_pucData[$iByteNow]) | $ucDataMask );

            }
            $ucSetMask = ( ($ucSetMask >> 1) & 0x7f );
        }

        $this->m_iDataBitNumber += $iBitNumber;
        
        /*
        {
            for( $i=0; $i<$iByteNeed; $i++ )
            {
                printf( "%02X ", ord($this->m_pucData[$i]) );
            }
            printf( "<br>\n" );
        }
        //*/

        return $bRet;
    }

    ///////////////////////////////////////////////////////////
    ///		２バイトデータを出力する
    ///////////////////////////////////////////////////////////
    //bool AddDataBits_US( unsigned short usData , int & iErrorType );
    protected function AddDataBits_US( $usData , &$iErrorType )
    {
        $bRet = TRUE;
        
        $ucHigh = ( usData >> 8 ) & 0xff;
        $ucLow  = usData & 0xff;

        $bRet = $this->AddDataBits( $ucHigh , 8 , $iErrorType );
        if( bRet )
        {
            $bRet = $this->AddDataBits( $ucLow , 8 , $iErrorType );
        }
        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		数字データのセット処理
    ///////////////////////////////////////////////////////////
	//bool AddDataBits_Numeric( const char * pcszData , int iDataLength , int & iErrorType );
    protected function AddDataBits_Numeric( $pcszData , $iDataLength , &$iErrorType )
    {
        global $c_typeMatrix;
        $bRet = TRUE;
        
        /*For Debug
        printf( "AddDataBits_Numeric..pcszData:%s, iDataLength=%d<br>\n", $pcszData, $iDataLength );
        //*/

        //データが数字のみかどうかのチェック
        /*for( int i=0 ; i<iDataLength ; i++ )
        {
            char cTemp = pcszData[i];

            //数字でない場合は即戻り
            if( !isdigit( cTemp ) )
            {
                iErrorType = QRERR_DATAISNOT_NUMERIC;
                return false;
            }
        }*/
        $checkString = substr( $pcszData, 0, $iDataLength );
        if( !ctype_digit( $checkString ) )
        {
            $iErrorType = QRERR_DATAISNOT_NUMERIC;
            return FALSE;
        }

        //数字の指示子出力
        $bRet = $this->AddDataBits( 0x01 , 4 , $iErrorType );

        //文字数指示子の出力
        $sOutputNumber = $iDataLength;

        $iModelNumber = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber;

        //１～９型の場合（１０ビット）
        if( $iModelNumber <= 9 )
        {
            $bRet = $this->AddDataBits( ($sOutputNumber >> 8) & 0x03  , 2 , $iErrorType );
        }

        //１０～２６型の場合（１２ビット）
        else if($iModelNumber <= 26 )
        {
            $bRet = $this->AddDataBits( ($sOutputNumber >> 8) & 0x0f , 4 , $iErrorType );
        }

        //２７～４０型の場合（１４ビット）
        else
        {
            $bRet = $this->AddDataBits( ($sOutputNumber >> 8) & 0x3f , 6 , $iErrorType );
        }
        
        //下位８ビット出力
        if( $bRet )
        {
            $bRet = $this->AddDataBits( $sOutputNumber & 0xff  , 8 , $iErrorType );
        }

        for( $i=0 ; $i<$iDataLength && $bRet ; $i+= 3 )
        {
            /*char szTemp[4];
            int iIndex = 0;*/
            $szTemp = "";
            $iIndex = 0;

            for( $j=$i ; $j<$iDataLength && $iIndex < 3 ; $j++,$iIndex++ )
            {
                $szTemp .= $pcszData[$j];
            }
            
            /* ForDebug
            printf( "szTemp=%s<br>\n", $szTemp );
            //*/

            $sTemp = intval( $szTemp );
            $sizeLength = strlen( $szTemp );

            //１文字のみ
            if( $sizeLength == 1 )
            {
                $bRet = $this->AddDataBits( $sTemp & 0x0f , 4 , $iErrorType );
            }
            //２文字
            else if( $sizeLength == 2 )
            {
                $bRet = $this->AddDataBits( $sTemp & 0x7f , 7 , $iErrorType );
            }
            //３文字
            else if( $sizeLength == 3 )
            {
                $bRet = $this->AddDataBits( ($sTemp >> 8) & 0x03 , 2 , $iErrorType );
                if( $bRet )
                {
                    $bRet = $this->AddDataBits( $sTemp & 0xff , 8 , $iErrorType );
                }
            }
        }

        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		アルファベットデータのセット処理
    ///////////////////////////////////////////////////////////
	//bool AddDataBits_Alphabet( const char * pcszData , int iDataLength , int & iErrorType );
    protected function AddDataBits_Alphabet( $pcszData , $iDataLength , &$iErrorType )
    {
        global $c_typeMatrix;
        $bRet = TRUE;

        $iAlphabetsLength = strlen( cszQrAlphabets );
        //データが英数字のみかどうかをチェックする
        for( $i=0 ; $i<$iDataLength ; $i++ )
        {
            $cTemp = $pcszData[$i];
            $bRet = FALSE;
            
            for( $j = 0 ; $j<$iAlphabetsLength ; $j++ )
            {
                if( $cTemp == substr( cszQrAlphabets, $j, 1) )
                {
                    $bRet = TRUE;
                    break;
                }
            }

            if( !$bRet )
            {
                return FALSE;
            }
        }

        //アルファベットの指示子出力
        $bRet = $this->AddDataBits( 0x02 , 4 , $iErrorType );

        //文字数の出力
        $usLength = $iDataLength;
        $iModelTypeNumber = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber;

        //１～９型
        if( $iModelTypeNumber <= 9 )
        {
            $bRet = $this->AddDataBits( ( $usLength >> 8 ) & 0x01 , 1 , $iErrorType );
        }
        //１０～２６型
        else if( $iModelTypeNumber <= 26 )
        {
            $bRet = $this->AddDataBits( ( $usLength >> 8 ) & 0x07 , 3 , $iErrorType );
        }
        //２７～４０型
        else
        {
            $bRet = $this->AddDataBits( ( $usLength >> 8 ) & 0x1f , 5 , $iErrorType );
        }

        if( $bRet )
        {
            $bRet = $this->AddDataBits( $usLength & 0xff , 8 , $iErrorType );
        }

        //２文字ずつ１１ビットに変換する
        for( $i=0 ; $i<$iDataLength && $bRet ; $i+=2 )
        {
            $usChar = 0;
            $iChar = 0;

            for( $iChar = 0; $iChar < 2 && $i+$iChar<$iDataLength ; $iChar++ )
            {
                $usChar *= 45;

                $alphabetsNumber = strlen(cszQrAlphabets);
                for( $j=0 ; $j < $alphabetsNumber ; $j++ )
                {
                    if( substr( cszQrAlphabets, $j, 1 ) == $pcszData[$i+$iChar] )
                    {
                        $usChar += $j;
                    }
                }
            }

            //１文字のみの出力の場合（６bit出力）
            if( $iChar == 1 )
            {
                $bRet = $this->AddDataBits( $usChar & 0x3f , 6 , $iErrorType );
            }
            //２文字の出力の場合（１１bit出力）
            else if( $iChar == 2 )
            {
                $bRet = $this->AddDataBits( ($usChar >> 8) & 0x07 , 3 , $iErrorType );
                if( $bRet )
                {
                    $bRet = $this->AddDataBits( $usChar & 0xff , 8 , $iErrorType );
                }
            }
        }

        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		アスキーデータのセット処理
    ///////////////////////////////////////////////////////////
	//bool AddDataBits_Ascii( const char * pcszData , int iDataLength , int & iErrorType );
    protected function AddDataBits_Ascii( $pcszData , $iDataLength , &$iErrorType )
    {
        global $c_typeMatrix;
        $bRet = TRUE;

        //アスキーの指示子出力
        $bRet = $this->AddDataBits( 0x04 , 4 , $iErrorType );

        if( $bRet )
        {
            //データ長の出力
            $iModelNumber = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber;

            //１型～９型
            if( $iModelNumber <= 9 )
            {
                $bRet = $this->AddDataBits( $iDataLength , 8 , $iErrorType );
            }
            //１０型～４０型
            else
            {
                $bRet = $this->AddDataBits_US( $iDataLength , $iErrorType );
            }
        }

        for( $i=0 ; $i<$iDataLength && $bRet ; $i++ )
        {
            $ucTemp = $pcszData[$i];
            $bRet = $this->AddDataBits( ord($ucTemp) , 8 , $iErrorType );
        }

        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		漢字データのセット処理
    ///////////////////////////////////////////////////////////
	//bool AddDataBits_Kanji( const char * pcszData , int iDataLength , int & iErrorType );
    protected function AddDataBits_Kanji( $pcszData , $iDataLength , &$iErrorType )
    {
        global $c_typeMatrix;
        
        $bRet = TRUE;

        //漢字かどうかのチェック
        for( $i=0 ; $i<$iDataLength ; $i+=2 )
        {
            $usTemp = ((( ord($pcszData[$i])) << 8) & 0xff00) |
                                        ((ord($pcszData[$i+1]) & 0x00ff) );
            
            //0x8140～0x9ffc
            if( !(0x8140 <= $usTemp && $usTemp <= 0x9ffc || 
                  0xe040 <= $usTemp && $usTemp <= 0xebbf) )
            {
                $iErrorType = QRERR_DATAISNOT_KANJI;
                return FALSE;
            }

        }

        //漢字モード
        $bRet = $this->AddDataBits( 0x08 , 4 , $iErrorType );
        if( !$bRet )
        {
            return $bRet;
        }

        //文字数の出力
        $sLength = floor($iDataLength / 2);
        $iModelNumber = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iModelNumber;

        //１～９型
        if( $iModelNumber <= 9 )
        {
            $bRet = $this->AddDataBits( ( $sLength & 0xff ) , 8 , $iErrorType );
        }
        else
        {
            //１０～２６型
            if( $iModelNumber <= 26 )
            {
                $bRet = $this->AddDataBits( (($sLength >> 8) & 0x03) , 2 , $iErrorType );
            }
            //２７～４０型
            else
            {
                $bRet = $this->AddDataBits( (($sLength >> 8) & 0x0f) , 4 , $iErrorType );
            }

            if( $bRet )
            {
                $bRet = $this->AddDataBits( ($sLength & 0xff) , 8 , $iErrorType );
            }
        }

        //文字の出力
        for( $i=0 ; $i<$iDataLength && $bRet ; $i+=2 )
        {
            $usCharacter = ((( ord($pcszData[$i])) << 8) & 0xff00) |
                                            ( ord($pcszData[$i+1]) & 0x00ff );

            //0x8140～0x9ffc
            if( 0x8140 <= $usCharacter && $usCharacter <= 0x9ffc )
            {
                //0x8140を引く
                $usCharacter -= 0x8140;
            }
            //0xe040～0xebbf
            else
            {
                //0xc140を引く
                $usCharacter -= 0xc140;
            }
            $usCharacter = ((($usCharacter >> 8) & 0xff) * 0xc0)	//上位バイトにC0を乗じる
                        +  (($usCharacter) & 0xff);				//下位バイトを足す

            //データの出力処理
            $bRet = $this->AddDataBits( ($usCharacter >> 8) & 0x1f , 5 , $iErrorType );
            if( $bRet )
            {
                $bRet = $this->AddDataBits( ($usCharacter & 0xff) , 8 , $iErrorType );
            }
        }

        return $bRet;
    }

    ///////////////////////////////////////////////////////////
    ///		データビットパターンを出力する
    ///////////////////////////////////////////////////////////
    //bool OutputDataBits( int & iErrorType );
    protected function OutputDataBits( &$iErrorType )
    {
        global $c_typeMatrix;
        
        $iColumnIndex = $this->m_iModuleNumber - 1;
        $iRowIndex    = $this->m_iModuleNumber - 1;

        $bDirectionUp = TRUE;
        $bFirstColumn = TRUE;

        $iRSBlockNumber = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iRSBlockNumber;
        $iDataByteNumber = floor( $c_typeMatrix[$this->m_iModelTypeIndex]->m_iDataBitNumber / 8 );
        $iBitIndex = 0;
        $iFirstCount = 0;   $iSecondCount = 0;
        $iFirstDataNumber = 0; $iSecondDataNumber = 0;

        $this->SeparateTotalNumber( $iRSBlockNumber , $iDataByteNumber 
                            , $iFirstCount , $iFirstDataNumber 
                            , $iSecondCount , $iSecondDataNumber );

        $bRet = TRUE;

        //ブロックごとのデータループ
        for( $iBlockIndex=0 ; $iBlockIndex<$iFirstDataNumber+1 ; $iBlockIndex++ )
        {
            //複数ブロックの出力
            for( $iSeq = 0 ; $iSeq < $iRSBlockNumber ; $iSeq++ )
            {
                $iByteIndex = 0;

                //ファーストカウント内の場合
                if( $iSeq < $iFirstCount )
                {
                    $iByteIndex = $iFirstDataNumber * $iSeq;

                    //ブロック内インデックスがデータ領域よりも大きい場合
                    if( $iBlockIndex >= $iFirstDataNumber )
                    {
                        continue;
                    }
                }
                else
                {
                    $iByteIndex = $iFirstDataNumber * $iFirstCount + ($iSeq - $iFirstCount) * $iSecondDataNumber;
                }

                $iByteIndex += $iBlockIndex;
                $iBitIndex = $iByteIndex * 8;

                for( $i=0 ; $i<8 ; $i++ )
                {
                    $this->GetNextDataPutPoint( $iColumnIndex , $iRowIndex , $bFirstColumn , $bDirectionUp );

                    if( $this->GetDataBit( $iBitIndex , TRUE ) )
                    {
                        $bRet = $this->SetModulePoint( $iColumnIndex , $iRowIndex , QR_DATA_DARK );
                    }
                    else
                    {
                        $bRet = $this->SetModulePoint( $iColumnIndex , $iRowIndex , QR_DATA_LIGHT );
                    }

                    $iBitIndex++;
                }
            }
        }

        $iErrorByteNumber = floor( ( $c_typeMatrix[$this->m_iModelTypeIndex]->m_iTotalBits - 
                              $c_typeMatrix[$this->m_iModelTypeIndex]->m_iDataBitNumber) / 8 / $iRSBlockNumber );
        //ブロックごとのデータのループ
        for( $iBlockIndex=0 ; $iBlockIndex < $iErrorByteNumber ; $iBlockIndex++ )
        {
            //ＲＳブロックごとのデータのループ
            for( $iSeq = 0 ; $iSeq < $iRSBlockNumber ; $iSeq++ )
            {
                $iByteIndex = $iSeq * $iErrorByteNumber + $iBlockIndex;
                $iBitIndex = $iByteIndex * 8;

                for( $i=0 ; $i<8 ; $i++ )
                {
                    $this->GetNextDataPutPoint( $iColumnIndex , $iRowIndex , $bFirstColumn , $bDirectionUp );

                    if( $this->GetDataBit( $iBitIndex , FALSE ) )
                    {
                        $bRet = $this->SetModulePoint( $iColumnIndex , $iRowIndex , QR_COLLECT_DARK );
                    }
                    else
                    {
                        $bRet = $this->SetModulePoint( $iColumnIndex , $iRowIndex , QR_COLLECT_LIGHT );
                    }

                    $iBitIndex++;
                }
            }
        }

        return $bRet;
    }

	//unsigned char GetDataBit( int iBitIndex , bool bDataRequest );
    ///////////////////////////////////////////////////////////
    ///		指定されたインデックスのビットを取得する
    ///		戻り値：ON：0x01
    ///				OFF：0x00
    ///////////////////////////////////////////////////////////
    protected function GetDataBit( $iBitIndex , $bDataRequest )
    {
        $iByteIndex = $iBitIndex / 8;

        //データのインデックスが負値の場合
        if( $iByteIndex < 0 )
        {
            return 0x00;
        }

        //マスクの取得
        $ucMask = $this->GetBitMask( 7 - $iBitIndex % 8 );

        //データがリクエストされている場合
        if( $bDataRequest )
        {
            if( $iByteIndex < $this->m_iDataSizeNumber && (ord($this->m_pucData[$iByteIndex]) & $ucMask) != 0 )
            {
                return 0x01;
            }
        }
        //誤り訂正がリクエストされている場合
        else
        {
            if( $iByteIndex < $this->m_iErrorSizeNumber && (ord($this->m_pucError[$iByteIndex]) & $ucMask) != 0 )
            {
                return 0x01;
            }
        }

        return 0x00;
    }
    
    /////////////////////////////////////////////////////////////////////
    ///		次のデータのセットポイントを取得する
    /////////////////////////////////////////////////////////////////////
	//bool GetNextDataPutPoint( int & iColumnIndex , int & iRowIndex
    //                          , bool & bFirstColumn
    //                          , bool & bDirectionUp );
    protected function GetNextDataPutPoint( &$iColumnIndex , &$iRowIndex
                                        , &$bFirstColumn
                                        , &$bDirectionUp )
    {
        //iColumnIndexおよびiRowIndexの位置決め
        while( $this->GetModulePoint( $iColumnIndex , $iRowIndex ) != QR_UNKNOWN )
        {
            $bDirectionChanged = FALSE;
            //２列の最初のデータの場合
            if( $bFirstColumn )
            {
                $iColumnIndex--;
                $bFirstColumn = FALSE;
            }
            //縦方向を変更する場合
            else
            {
                //上に移動する場合
                if( $bDirectionUp )
                {
                    //上の行に
                    $iRowIndex--;

                    //最上段を超えた場合
                    if( $iRowIndex < 0 )
                    {
                        $iRowIndex = 0;
                        $iColumnIndex--;
                        $bDirectionUp = FALSE;
                    }
                    else
                    {
                        $iColumnIndex++;
                    }
                }
                //下に移動する場合
                else
                {
                    //下の行に
                    $iRowIndex++;

                    //最下段を超えた場合
                    if( $iRowIndex >= $this->m_iModuleNumber )
                    {
                        $iRowIndex = $this->m_iModuleNumber - 1;
                        $iColumnIndex--;
                        $bDirectionUp = TRUE;
                    }
                    else
                    {
                        $iColumnIndex++;
                    }
                }

                //特殊（列が、タイミングパターン（縦の列の場合）
                if( $iColumnIndex == 6 )
                {
                    $iColumnIndex = 5;
                }

                $bFirstColumn = TRUE;
            }
        }
        /* ForDebug
        printf( "Result of GetNextDataPutPoint iColumnIndex=%d, iRowIndex=%d, bFirstColumn=%s, bDirectionUp = %s<br>\n"
               , $iColumnIndex, $iRowIndex
               , ($bFirstColumn)?("true"):("false"), ($bDirectionUp)?("true"):("false") );
        //*/
        return TRUE;
    }

    ///////////////////////////////////////////////////////////
    ///		マスクパターンの決定
    ///////////////////////////////////////////////////////////
	//bool DecideMaskPattern( int & iErrorType );
    protected function DecideMaskPattern( &$iErrorType )
    {
        $iMinLostPoint;
        $iMinQRMaskPattern;
        $bRet = TRUE;

        for( $this->m_iQRMaskPattern = 0 ; $this->m_iQRMaskPattern < 8 ; $this->m_iQRMaskPattern++ )
        {
            $iLostPoint = 0;

            $iBlackPointNumber = 0;

            //列ごとのチェック
            for( $iRowIndex = 0 ; $iRowIndex < $this->m_iModuleNumber ; $iRowIndex++ )
            {
                $iSeqBlackModuleNumber = 0;
                $iSeqWhiteModuleNumber = 0;

                for( $iColumnIndex = 0 ; $iColumnIndex < $this->m_iModuleNumber ; $iColumnIndex++ )
                {
                    //暗ポイントの場合
                    if( $this->IsDarkPoint( $iColumnIndex , $iRowIndex ) )
                    {
                        $iSeqBlackModuleNumber++;
                        $iBlackPointNumber++;

                        //連続明ポイントのチェック
                        $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqWhiteModuleNumber );
                        $iSeqWhiteModuleNumber = 0;
                    }
                    //明ポイントの場合
                    else
                    {
                        $iSeqWhiteModuleNumber++;

                        //連続暗ポイントのチェック
                        $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqBlackModuleNumber );
                        $iSeqBlackModuleNumber = 0;
                    }

                    //２×２ブロックの失点カウント
                    $iLostPoint += $this->GetLostPoint_ModuleBlock( $iColumnIndex , $iRowIndex );

                    //1：1：3：1：1比率のチェック
                    $iLostPoint += $this->GetLostPoint_PositionDitect( $iColumnIndex , $iRowIndex );
                }

                //連続暗ポイントのチェック
                $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqBlackModuleNumber );
                //連続明ポイントのチェック
                $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqWhiteModuleNumber );
            }

            //暗／明比率
            $iLostPoint += $this->GetLostPoint_BlackWhiteRatio( $iBlackPointNumber );
            /*
            printf( "%d iLostPoint = %d<br>\n", $this->m_iQRMaskPattern, $iLostPoint );
            //*/

            //行ごとのチェック
            for( $iColumnIndex = 0 ; $iColumnIndex < $this->m_iModuleNumber ; $iColumnIndex++ )
            {
                $iSeqBlackModuleNumber = 0;
                $iSeqWhiteModuleNumber = 0;

                //列ごとのチェック
                for( $iRowIndex = 0 ; $iRowIndex < $this->m_iModuleNumber ; $iRowIndex++ )
                {
                    //暗ポイントの場合
                    if( $this->IsDarkPoint( $iColumnIndex , $iRowIndex ) )
                    {
                        $iSeqBlackModuleNumber++;

                        //連続明ポイントのチェック
                        $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqWhiteModuleNumber );
                        $iSeqWhiteModuleNumber = 0;
                    }
                    //明ポイントの場合
                    else
                    {
                        $iSeqWhiteModuleNumber++;

                        //連続暗ポイントのチェック
                        $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqBlackModuleNumber );
                        $iSeqBlackModuleNumber = 0;
                    }
                }

                //連続暗ポイントのチェック
                $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqBlackModuleNumber );
                //連続明ポイントのチェック
                $iLostPoint += $this->GetLostPoint_SeqModule( $iSeqWhiteModuleNumber );
            }
            
            /*
            printf( "%d iLostPoint = %d<br>\n", $this->m_iQRMaskPattern, $iLostPoint );
            //*/

            //一番失点の少ないマスクパターンの取得
            if( $this->m_iQRMaskPattern == 0 )
            {
                $iMinLostPoint = $iLostPoint;
                $iMinQRMaskPattern = 0;
            }
            else if( $iMinLostPoint > $iLostPoint )
            {
                $iMinLostPoint = $iLostPoint;
                $iMinQRMaskPattern = $this->m_iQRMaskPattern;
            }
        }

        $this->m_iQRMaskPattern = $iMinQRMaskPattern;

        return $bRet;
    }

    ///////////////////////////////////////////////////////////
    ///		失点のカウント（連続暗・明）
    ///////////////////////////////////////////////////////////
	//int GetLostPoint_SeqModule( int iSeqBlock );
    protected function GetLostPoint_SeqModule( $iSeqBlock )
    {
        if( $iSeqBlock < 5 )
        {
            return 0;
        }

        return ($iSeqBlock - 5) + 3;
    }
    
    ///////////////////////////////////////////////////////////
    ///		失点のカウント（指定列・行を開始点とした２×２ブロック）
    ///////////////////////////////////////////////////////////
	//int GetLostPoint_ModuleBlock( int iColumnIndex , int iRowIndex );
    protected function GetLostPoint_ModuleBlock( $iColumnIndex , $iRowIndex )
    {
        if( $iColumnIndex >= $this->m_iModuleNumber - 1 || $iRowIndex >= $this->m_iModuleNumber - 1 )
        {
            return 0;
        }

        $bDark = $this->IsDarkPoint( $iColumnIndex , $iRowIndex );
        if( $bDark == $this->IsDarkPoint( $iColumnIndex + 1 , $iRowIndex     ) &&
            $bDark == $this->IsDarkPoint( $iColumnIndex     , $iRowIndex + 1 ) &&
            $bDark == $this->IsDarkPoint( $iColumnIndex + 1 , $iRowIndex + 1 ) )
        {
            return 3;
        }
        return 0;
    }

    /////////////////////////////////////////////////////////////
    ///		失点のカウント（指定位置から1：1：3：1：1の行列チェック
    ///						マッチする場合には前後の明チェック）
    /////////////////////////////////////////////////////////////
	//int GetLostPoint_PositionDitect( int iColumnIndex , int iRowIndex );
    protected function GetLostPoint_PositionDitect( $iColumnIndex , $iRowIndex )
    {
        $bDirectionColumn = TRUE;
        $iLostPoint = 0;

        //縦横ループ
        for( $i=0 ; $i<2 ; $i++ )
        {
            if( $i != 0 )
            {
                $bDirectionColumn = FALSE;
            }
            
            //調査する方向に７ドット以上のデータが存在しない場合
            if( ($bDirectionColumn && $iColumnIndex >= $this->m_iModuleNumber - 7) ||
                (!$bDirectionColumn && $iRowIndex >= $this->m_iModuleNumber - 7 ) )
            {
                continue;
            }

            //1：1：3：1：1のチェック
            $bMatched = TRUE;

            for( $iIndex = 0 ; $iIndex<7 ; $iIndex++ )
            {
                $bNeedDark = FALSE;
                if( $iIndex == 3 || $iIndex % 2 == 0 )
                {
                    $bNeedDark = TRUE;
                }

                if( $this->IsDarkPoint( $iColumnIndex + (($bDirectionColumn)?($iIndex):(0))
                                ,$iRowIndex    + (($bDirectionColumn)?(0):($iIndex)) ) != $bNeedDark )
                {
                    $bMatched = FALSE;
                    break;
                }
            }

            //1：1：3：1：1が存在しない場合
            if( !$bMatched )
            {
                continue;
            }

            //前のスペースをカウントする
            if( ( $bDirectionColumn && $iColumnIndex > 3) || 
                (!$bDirectionColumn && $iRowIndex > 3 ) )
            {
                $bMatched = TRUE;

                for( $iIndex = 0 ; $iIndex < 4 ; $iIndex++ )
                {
                    $iPlusCount = -$iIndex - 1;
                    if( $this->IsDarkPoint( $iColumnIndex + (($bDirectionColumn)?($iPlusCount):(0))
                                    ,$iRowIndex    + (($bDirectionColumn)?(0):($iPlusCount)) ) )
                    {
                        $bMatched = FALSE;
                        break;
                    }
                }

                //前４つが明の場合
                if( $bMatched )
                {
                    $iLostPoint += 40;
                    continue;
                }
            }

            //後ろのスペースをカウントする
            if( ( $bDirectionColumn && $iColumnIndex < $this->m_iModuleNumber - 7 - 4 ) ||
                (!$bDirectionColumn && $iRowIndex    < $this->m_iModuleNumber - 7 - 4 ) )
            {
                $bMatched = TRUE;

                for( $iIndex = 0 ; $iIndex < 4 ; $iIndex++ )
                {
                    $iPlusCount = $iIndex + 7 + $i;
                    if( $this->IsDarkPoint( $iColumnIndex + (($bDirectionColumn)?($iPlusCount):(0))
                                           ,$iRowIndex    + (($bDirectionColumn)?(0):($iPlusCount)) ) )
                    {
                        $bMatched = FALSE;
                        break;
                    }
                }

                //後ろ４つが明の場合
                if( $bMatched )
                {
                    $iLostPoint += 40;
                }
            }

        }
        
        /*
        if( $iLostPoint != 0 )
        {
            printf( "%d GetLostPoint_PositionDitect iLostPoint = %d<br>\n", $s_iCount, $iLostPoint );
        }
        $s_iCount++;
        //*/
        
        return $iLostPoint;
    }

    ///////////////////////////////////////////////////////////
    ///		失点のカウント（暗・明比率）
    ///////////////////////////////////////////////////////////
	//int GetLostPoint_BlackWhiteRatio( int iBlackPoint );
    protected function GetLostPoint_BlackWhiteRatio( $iBlackPoint )
    {
        $iAllNumber = $this->m_iModuleNumber * $this->m_iModuleNumber;
        $iMod = floor($iAllNumber / 2) - $iBlackPoint;
        if( $iMod < 0 )
        {
            $iMod = -$iMod;
        }

        return floor( floor( ($iMod * 100 + $iAllNumber - 1) / $iAllNumber) / 5 ) * 10;
    }

    ///////////////////////////////////////////////////////////
    ///		誤り訂正コード語の作成処理
    ///////////////////////////////////////////////////////////
	//bool CreateErrorCollectData( int & iErrorType );
    protected function CreateErrorCollectData( &$iErrorType )
    {
        global $c_typeMatrix;
        
        $bRet = TRUE;

        $this->ReleaseError();

        //誤りデータコード語数（ビット数）
        $iErrorBit = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iTotalBits 
                      - $c_typeMatrix[$this->m_iModelTypeIndex]->m_iDataBitNumber;
        $this->m_iErrorSizeNumber = $iErrorBit / 8;
        $m_pucError = "";//new unsigned char[m_iErrorSizeNumber];

        $iFirstCount = 0 ; $iSecondCount = 0;
        $iFirstDataNumber = 0; $iSecondDataNumber = 0;

        $iSeparateCount = $c_typeMatrix[$this->m_iModelTypeIndex]->m_iRSBlockNumber;
        $iTotalNumber = floor($c_typeMatrix[$this->m_iModelTypeIndex]->m_iDataBitNumber / 8);

        $this->SeparateTotalNumber( $iSeparateCount , $iTotalNumber , $iFirstCount , $iFirstDataNumber , $iSecondCount , $iSecondDataNumber );

        $iDataIndex = 0;
        $iErrorIndex = 0;
        $iErrorNumber = floor( ( $c_typeMatrix[$this->m_iModelTypeIndex]->m_iTotalBits
                            -$c_typeMatrix[$this->m_iModelTypeIndex]->m_iDataBitNumber) / 8 / $iSeparateCount );

        //各ブロックごとにリードソロモンによる符号化を実行する
        for( $i=0 ; $i<$iSeparateCount ; $i++ )
        {
            $iDataNumber = $iSecondDataNumber;
            if( $i < $iFirstCount )
            {
                $iDataNumber = $iFirstDataNumber;
            }

            $qrReadSolomon = new CQRReadSolomon();

            /* For debug
            printf( "Before read solomon<br>\n" );
            printf( "this->m_pucData = " );
            for( $i=0; $i<$iDataNumber; $i++ )
            {
                printf( "%02X ", ord($this->m_pucData[$i]) );
            }
            printf( "<br>\n" );
            printf( "this->m_pucError = " );
            for( $i=0; $i<$iErrorNumber; $i++ )
            {
                printf( "%02X ", ord( $this->m_pucError[$i] ) );
            }
            printf( "<br>\n" );
            //*/
            
            $qrReadSolomon->CodeTo( $this->m_pucData, $iDataIndex , $iDataNumber , $this->m_pucError , $iErrorIndex , $iErrorNumber );

            /* For debug
            printf( "After read solomon<br>\n" );
            printf( "this->m_pucData = " );
            for( $i=0; $i<$iDataNumber; $i++ )
            {
                printf( "%02X ", ord($this->m_pucData[$i]) );
            }
            printf( "<br>\n" );
            printf( "this->m_pucError = " );
            for( $i=0; $i<$iErrorNumber; $i++ )
            {
                printf( "%02X ", ord( $this->m_pucError[$i] ) );
            }
            printf( "<br>\n" );
            //*/

            $iDataIndex += $iDataNumber;
            $iErrorIndex += $iErrorNumber;
        }
        return $bRet;
    }

    ///////////////////////////////////////////////////////////
    ///		データの分割
    ///		端数が発生する場合、iFirstCountとiSecondCountに分離して
    ///		整数で分割できるように調整する
    ///		RSブロックに対するデータ数を計算するのに使用する
    ///////////////////////////////////////////////////////////
	//bool SeparateTotalNumber( int iSeparateCount , int iTotalNumber
	//                         , int & iFirstCount , int & iFirstNumber
    //                          , int & iSecondCount , int & iSecondNumber );
    protected function SeparateTotalNumber( $iSeparateCount , $iTotalNumber
                                        , &$iFirstCount , &$iFirstNumber
                                        , &$iSecondCount , &$iSecondNumber )
    {
        $iFirstNumber = floor($iTotalNumber / $iSeparateCount);

        $iSecondCount = $iTotalNumber - $iFirstNumber * $iSeparateCount;
        if( $iSecondCount == 0 )
        {
            $iSecondNumber = 0;
        }
        else
        {
            $iSecondNumber = $iFirstNumber + 1;
        }

        $iFirstCount = $iSeparateCount - $iSecondCount;

        return TRUE;
    }
    
	///////////////////////////////////////////////////////////
	///		指定された行列に対応するポイントデータをセットする
	///////////////////////////////////////////////////////////
	protected function SetModulePoint( $iColumnIndex , $iRowIndex , $iQRPoint )
	{
		if( $iColumnIndex < 0 || $iColumnIndex >= $this->m_iModuleNumber ||
			$iRowIndex    < 0 || $iRowIndex    >= $this->m_iModuleNumber )
		{
			return FALSE;
		}
		$this->m_ppucQRMatrix[$iColumnIndex][$iRowIndex] = $iQRPoint;

		return TRUE;
	}

	///////////////////////////////////////////////////////////
	///		指定された行列に対応するポイントデータを取得する
	///////////////////////////////////////////////////////////
	protected function GetModulePoint( $iColumnIndex , $iRowIndex )
	{
		if( $iColumnIndex < 0 || $iColumnIndex >= $this->m_iModuleNumber ||
			$iRowIndex    < 0 || $iRowIndex    >= $this->m_iModuleNumber )
		{
			return QR_UNKNOWN;
		}

		return $this->m_ppucQRMatrix[$iColumnIndex][$iRowIndex];
	}

	///////////////////////////////////////////////////////////
	///		指定行、列のマスクパターンを取得する
	///////////////////////////////////////////////////////////
	protected function GetMaskPoint( $iColumnIndex , $iRowIndex , $iQRMaskPattern = -1)
	{
		if( $iQRMaskPattern < 0 )
		{
			$iQRMaskPattern = $this->m_iQRMaskPattern;
		}

		if( $iQRMaskPattern < 0 || $iQRMaskPattern > 7 )
		{
			return QR_UNKNOWN;
		}

		$bMask = TRUE;

		//仕様書がi,jなので、混乱しないように変換
		$i = $iRowIndex;
		$j = $iColumnIndex;

		switch( $iQRMaskPattern )
		{
		case 0:	$bMask = (($i + $j) % 2 == 0);							break;
		case 1:	$bMask = ($i % 2 == 0);								   break;
		case 2:	$bMask = ($j % 3 == 0);							    	break;
		case 3:	$bMask = (($i + $j) % 3 == 0);							break;
		case 4:	$bMask = ((floor($i / 2) + floor($j / 3)) % 2 == 0 );	break;
		case 5:	$bMask = (($i * $j) % 2 + ($i * $j) % 3 == 0);			break;
		case 6:	$bMask = ((($i * $j) % 2 + ($i * $j) % 3 ) % 2 == 0);	break;
		case 7:	$bMask = ((($i * $j) % 3 + ($i + $j) % 2 ) % 2 == 0);	break;
		}

		return ($bMask)?(QR_MASK_DARK):(QR_MASK_LIGHT);
	}
    
    ///////////////////////////////////////////////////////////
    ///		ビット多項式の剰余を求める
    ///		puszCoefficientsの長さはiErrorLength+1となる
    ///////////////////////////////////////////////////////////
    //bool CQRMatrix::GetExpressionMod( unsigned char *puszBaseData , int iBaseDataLength,
    //                           unsigned char *puszCoefficients ,
    //                           unsigned char *puszError , int iErrorLength )
    // $puszError = $puszBaseData[ $iErrorOffset ]として扱う
    function GetExpressionMod( &$puszBaseData, $iBaseDataLength, $puszCoefficients, $iErrorOffset, $iErrorLength )
    {
        /* ForDebug
        for( $i=0; $i<$iBaseDataLength; $i++ )
        {
            printf( "%02X ", ord( $puszBaseData[ $i ] ) );
        }
        printf( "<br>\n" );
        //*/
        
        $bRet = TRUE;
        $iCoefficientsLength = $iErrorLength + 1;

        $sizeDataLength = $iBaseDataLength + $iCoefficientsLength + 1;
        $puszData = "";//new unsigned char[sizeDataLength];

        //データ初期化
        memset( $puszData , "0" , $sizeDataLength - 1 );
        //puszData[sizeDataLength - 1] = '\0';

        memcpy( $puszData , $puszBaseData , $iBaseDataLength );
        /* For Debug
        for( $i=0; $i<$sizeDataLength - 1; $i++ )
        {
            printf( "%02X ", ord($puszData[ $i ]) );
        }
        printf( "<br>\n" );
        //*/
        
        for( $iKeta = 0 ; $iKeta < $iBaseDataLength ; $iKeta++ )
        {
            //printf( "%02X ", ord( $puszData[$iKeta] ) );
            
            if( $puszData[$iKeta] == "0" )
            {
                continue;
            }

            for( $i=0 ; $i<$iCoefficientsLength ; $i++ )
            {
                if( $puszData[$iKeta + $i] == $puszCoefficients[$i] )
                {
                    $puszData[$iKeta + $i] = "0";
                }
                else
                {
                    $puszData[$iKeta + $i] = "1";
                }
            }
        }

        /* For Debug
        for( $i=0; $i<$iErrorLength; $i++ )
        {
            printf( "%02X ", ord($puszData[ $iBaseDataLength + $i ]) );
        }
        printf( "<br>\n" );
        //*/
        memcpy_offset( $puszBaseData, $iErrorOffset , $puszData , $iBaseDataLength , $iErrorLength );

        return $bRet;
    }

}

?>
