<?php
require_once( dirname(__FILE__)."/QRReadSolomonInformation.php" );

//static int SortCompareNumberAlpha( const void * pcParam1 , const void * pcParam2 );
function QRSortCompareNumberAlpha( $pAlphaNumber1 , $pAlphaNumber2 )
{
	return $pAlphaNumber1->m_sNumber - $pAlphaNumber2->m_sNumber;
}

class CQRReadSolomon
{
	//struct SAlphaNumber * m_pNumberAlpha;
    protected $m_pNumberAlpha;
    
    public function __construct()
    {
        global $c_alphaNumber;
        
        $this->m_pNumberAlpha = array();
        for( $i=0 ; $i<256 ; $i++ )
        {
            $this->m_pNumberAlpha[] = $c_alphaNumber[$i];
        }
        usort( $this->m_pNumberAlpha , "QRSortCompareNumberAlpha");
        
        $this->m_pNumberAlpha[0]->m_sNumber = 0;
        $this->m_pNumberAlpha[0]->m_sAlpha = 0;

        $this->m_pNumberAlpha[1]->m_sNumber = 1;
        $this->m_pNumberAlpha[1]->m_sAlpha = 0;
        
        /* for debug
        for( $i=0; $i<256; $i++ )
        {
            printf( "m_sNumber = %d, m_sAlpha=%d<br>\n", $this->m_pNumberAlpha[$i]->m_sNumber
                                                       , $this->m_pNumberAlpha[$i]->m_sAlpha );
        }
        //*/
    }
    
    ///////////////////////////////////////////////////////////
    ///		リードソロモン符号化の実行
    ///		引数：
    ///			pucBase		：	(IN)	コード化を行うデータ列
    ///			iDataLength	：	(IN)	コード化を行うデータ数
    ///			pucCode		：	(OUT)	コード化結果の格納用配列
    ///			iCodeLength	：	(IN)	コード化結果の格納用配列要素数＝コード語数
    ///////////////////////////////////////////////////////////
    //bool CodeTo( unsigned char * pucBase , int iBaseLength
    //			,unsigned char * pucCode , int iCodeLength );    
    public function CodeTo( &$pucDataBase, $iBaseOffset, $iBaseLength
                           ,&$pucCodeBase, $iCodeOffset, $iCodeLength )
    //$pucBase = $pucDataBase[ $iBaseOffset ]
    //$pucCode = $pucCodeBase[ $iCodeOffset ] として扱う
    {
        global $c_alphaNumber;
        
        $bRet = TRUE;

        //生成多項式の係数を取得する
        $pucCoefficients = $this->GetExpressionCoefficients( $iCodeLength );
        //var_dump( $pucCoefficients );

        if( $pucCoefficients == NULL )
        {
            return FALSE;
        }

        //ベースデータの複製を作成する
        $iDataLength = $iBaseLength + $iCodeLength;
        $pucData = "";
        memset( $pucData , "\x00" , $iDataLength );
        memcpy_offset( $pucData, 0, $pucDataBase, $iBaseOffset , $iBaseLength );
        
        /* For Debug
        printf( "pucData = " );
        for( $i=0; $i<$iDataLength; $i++ )
        {
            printf( "%02X ", ord($pucData[$i]) );
        }
        printf( "<br>\n" );
        //*/

        $iDataLast = $iDataLength - $iCodeLength;
        for( $iDataIndex = 0; $iDataIndex < $iDataLast ; $iDataIndex++ )
        {
            //データ係数
            $sKeisu = ord($pucData[$iDataIndex]);
            
            //printf( "%d ", $sKeisu );

            //対象となる桁が既に0の場合
            if( !$sKeisu )
            {
                continue;
            }

            //係数に対するアルファの累乗を取得する
            $sAlpha = $this->m_pNumberAlpha[$sKeisu]->m_sAlpha ;
            //printf( "%d ", $sAlpha );

            //生成多項式の係数リストにα累乗を加算する（累乗同士の掛け算＝＞乗数の足し算）
            for( $i=0 ; $i<$iCodeLength+1 ; $i++ )
            {
                $sTemp = $pucCoefficients[$i] + $sAlpha;

                //α＾255==1より、桁を落とす
                while( $sTemp > 255 )
                {
                    $sTemp -= 255;
                }
                
                //各桁の排他的論理和を取得し、データビットを埋める
                $pucData[$iDataIndex + $i] = chr((ord($pucData[$iDataIndex + $i])) ^ ($c_alphaNumber[$sTemp]->m_sNumber));
            }
        }

        //剰余データを配列にセットする
        for( $i=0 ; $i<$iCodeLength ; $i++ )
        {
            $iDataIndex = $iDataLength - $i - 1;
            $iCodeIndex = $iCodeLength - $i - 1;

            $pucCodeBase[$iCodeOffset + $iCodeIndex] = $pucData[$iDataIndex];
        }

        return $bRet;
    }
    
    ///////////////////////////////////////////////////////////
    ///		生成多項式の係数を取得する
    ///		引数：
    ///			iCodeLength	：	(IN)	コード語数
    ///		戻り値：
    ///			係数αの累乗リスト（pucCoefficients）
    ///			x^iCodeLength      *α^(pucCoefficients[0])
    ///			+ x^(iCodeLength-1)*α^(pucCoefficients[1])
    ///			+ x^(iCodeLength-2)*α^(pucCoefficients[2])
    ///			+ …………
    ///			+ x                *α^(pucCoefficients[iCodeLength-1])
    ///			+ 1                *α^(pucCoefficients[iCodeLength  ])
    ///
    ///			※（^）は累乗を意味する。xorではないので念のため
    ///////////////////////////////////////////////////////////
	//unsigned char * GetExpressionCoefficients( int iCodeLength );
    protected function GetExpressionCoefficients( $iCodeLength )
    {
        global $c_readSolomonExpression;
        
        $iExpressionIndex = 0;
        for( ; $c_readSolomonExpression[$iExpressionIndex]->m_iCodeNumber != $iCodeLength && 
               $c_readSolomonExpression[$iExpressionIndex]->m_iCodeNumber >= 0
             ; $iExpressionIndex++ );

        //生成多項式内のリストに指定のコード語数データが存在しない場合
        if( $c_readSolomonExpression[$iExpressionIndex]->m_iCodeNumber != $iCodeLength )
        {
            return NULL;
        }

        //引数リストのループ
        /*size_t sizeLength = strlen( (c_readSolomonExpression + iExpressionIndex)->m_pcszExpression) + 1;
        char * pszExpression = new char[sizeLength];
    #ifdef WIN32
        strcpy_s( pszExpression , sizeLength , (c_readSolomonExpression + iExpressionIndex)->m_pcszExpression );
    #else
        strcpy( pszExpression , (c_readSolomonExpression + iExpressionIndex)->m_pcszExpression );
    #endif

        //カンマ区切りを１次元リストに変更する
        for( int i=0 ; i<(int)sizeLength ; i++ )
        {
            if( *(pszExpression + i) == ',' )
            {
                *(pszExpression + i) = '\0';
            }
        }*/
        $pszExpressions = explode( ",", $c_readSolomonExpression[$iExpressionIndex]->m_pszExpression );
        $iExpressionsNumber = count( $pszExpressions );
        
        $pucCoefficients = array();
        
        for( $iIndex=0; $iIndex<$iExpressionsNumber; $iIndex++ )
        {
            $pucCoefficients[$iIndex] = intval( $pszExpressions[$iIndex] );
        }

        return $pucCoefficients;
    }
}

?>